#
# Copyright (c) 2004-2011 Purdue University All rights reserved.
# 
# Developed by: HUBzero Technology Group, Purdue University
#               http://hubzero.org
# 
# HUBzero is free software: you can redistribute it and/or modify it under the terms of the
# GNU Lesser General Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
# 
# HUBzero is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.  You should have received a
# copy of the GNU Lesser General Public License along with HUBzero.
# If not, see <http://www.gnu.org/licenses/>.
# 
# GNU LESSER GENERAL PUBLIC LICENSE
# Version 3, 29 June 2007
# Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
#
import os.path
import re

from hubzero.submit.TimeConversion import minTohhmmss

class RemoteBatchSLURM:
   def __init__(self,
                localJobId,
                instanceId,
                appScriptName,
                environment,
                transferExecutable,
                executable,
                arguments,
                stdinput,
                isMultiCoreRequest,
                siteInfo,
                managerInfo,
                nNodes,
                ppn,
                wallTime,
                quotaLimit,
                timeHistoryLogs):
      self.localJobId               = localJobId
      self.instanceId               = instanceId
      self.appScriptName            = appScriptName
      self.environment              = environment
      self.stageFiles               = siteInfo['stageFiles']
      self.transferExecutable       = transferExecutable
      if transferExecutable:
         if self.stageFiles:
            self.executable         = os.path.join(".",os.path.basename(executable))
         else:
            self.executable         = executable
      else:
         if not executable.startswith('/') and not executable.startswith('$') and siteInfo['remoteApplicationRootDirectory'] != "":
            self.executable         = os.path.join(siteInfo['remoteApplicationRootDirectory'],executable)
         else:
            self.executable         = executable
      self.arguments                = arguments
      self.stdinput                 = stdinput
      self.isMultiCoreRequest       = isMultiCoreRequest
      self.computationMode          = managerInfo['computationMode']
      self.preManagerCommands       = managerInfo['preManagerCommands']
      self.managerCommand           = managerInfo['managerCommand']
      self.postManagerCommands      = managerInfo['postManagerCommands']
      self.nNodes                   = nNodes
      self.ppn                      = ppn
      self.remoteBatchPartition     = siteInfo['remoteBatchPartition']
      self.remoteBatchPartitionSize = siteInfo['remoteBatchPartitionSize']
      self.remoteBatchConstraints   = siteInfo['remoteBatchConstraints']
      self.wallTime                 = minTohhmmss(wallTime)
      self.quotaLimit               = quotaLimit
      self.timestampStart           = timeHistoryLogs['timestampStart']
      self.timestampFinish          = timeHistoryLogs['timestampFinish']
      self.timeResults              = timeHistoryLogs['timeResults']

      self.nodeFileName = ""
      self.nodeList     = []


   def __makeSerialTemplate(self):
      return """#!/bin/sh
PARTITION
CONSTRAINTS
#SBATCH --nodes=NNODES-NNODES
#SBATCH --ntasks-per-node=PPN
#SBATCH --time=WALLTIME
#SBATCH -J JOB_JOBID
#SBATCH -o slurm_JOBID.stdout
#SBATCH -e slurm_JOBID.stderr

ENVIRONMENT

# RemoteBatchSLURM:makeSerialTemplate

trap cleanup HUP INT QUIT ABRT TERM

cleanup()
{
   echo "Abnormal termination by signal"
   if [ ! -s TS_FINISH ] ; then
      date +"%s" > TS_FINISH
   fi
   exit 1
}

export PATH=${SLURM_SUBMIT_DIR}:${PATH}
# Change to directory where job was submitted.
cd ${SLURM_SUBMIT_DIR}

TIMEPATH=
for timePath in ${HOME}/bin/time /usr/bin/time /usr/local/bin/time /apps/submit/bin/time ; do
   if [ -x ${timePath} ] ; then
      TIMEPATH=${timePath}
      break
   fi
done

QUOTA

date +"%s" > TS_START

${TIMEPATH} -p -o TIME_RESULTS EXECUTABLE ARGUMENTS STDIN > JOBID.stdout 2> JOBID.stderr

date +"%s" > TS_FINISH
"""


   def __buildSerialFile(self):
      # setup regex's for the template
      re_stdinput    = re.compile("STDIN")
      re_wallTime    = re.compile("WALLTIME")
      re_partition   = re.compile("PARTITION")
      re_constraints = re.compile("CONSTRAINTS")
      re_environment = re.compile("ENVIRONMENT")
      re_executable  = re.compile("EXECUTABLE")
      re_arguments   = re.compile("ARGUMENTS")
      re_tsStart     = re.compile("TS_START")
      re_tsFinish    = re.compile("TS_FINISH")
      re_timeResults = re.compile("TIME_RESULTS")
      re_jobid       = re.compile("JOBID")
      re_nnodes      = re.compile("NNODES")
      re_ppn         = re.compile("PPN")
      re_quota       = re.compile("QUOTA")

      template = self.__makeSerialTemplate()

      partitionSBATCH = ""
      if self.remoteBatchPartition != "":
         partitionSBATCH = "#SBATCH --partition " + self.remoteBatchPartition

      constraintSBATCH = ""
      if self.remoteBatchConstraints != "":
         constraintSBATCH = "#SBATCH --constraint " + self.remoteBatchConstraints

      environmentExport = ""
      environmentVars = self.environment.split()
      for environmentVar in environmentVars:
         environmentExport += "export " + environmentVar + "\n"

      if self.stdinput == "":
         template = re_stdinput.sub(self.stdinput,template)
      else:
         template = re_stdinput.sub(" < "+self.stdinput,template)
      template = re_wallTime.sub(self.wallTime,template)
      template = re_partition.sub(partitionSBATCH,template)
      template = re_constraints.sub(constraintSBATCH,template)
      template = re_environment.sub(environmentExport,template)
      template = re_executable.sub(self.executable,template)
      template = re_arguments.sub(self.arguments,template)
      template = re_tsStart.sub(self.timestampStart,template)
      template = re_tsFinish.sub(self.timestampFinish,template)
      template = re_timeResults.sub(self.timeResults,template)
      template = re_jobid.sub("%s_%s" % (self.localJobId,self.instanceId),template)
      if self.remoteBatchPartitionSize != "":
         template = re_nnodes.sub(self.remoteBatchPartitionSize,template)
      else:
         template = re_nnodes.sub(self.nNodes,template)
      template = re_ppn.sub(self.ppn,template)
      if self.quotaLimit > 0:
         template = re_quota.sub("ulimit -f " + str(self.quotaLimit),template)
      else:
         template = re_quota.sub("",template)

      return(template)


   def __makeMPITemplate(self):
      return """#!/bin/sh
PARTITION
CONSTRAINTS
#SBATCH --nodes=NNODES-NNODES
#SBATCH --ntasks-per-node=PPN
#SBATCH --time=WALLTIME
#SBATCH -J JOB_JOBID
#SBATCH -o slurm_JOBID.stdout
#SBATCH -e slurm_JOBID.stderr

ENVIRONMENT

# RemoteBatchSLURM:makeMPITemplate

trap cleanup HUP INT QUIT ABRT TERM

cleanup()
{
   echo "Abnormal termination by signal"
   if [ ! -s TS_FINISH ] ; then
      date +"%s" > TS_FINISH
   fi
   exit 1
}

export PATH=${SLURM_SUBMIT_DIR}:${PATH}
# Change to directory where job was submitted.
cd ${SLURM_SUBMIT_DIR}

chmod +x APPSCRIPTNAME
date +"%s" > TS_START

if [ "x${SLURM_NPROCS}" = "x" ] ; then
   if [ "x${SLURM_NTASKS_PER_NODE}" = "x" ] ; then
      SLURM_NTASKS_PER_NODE=1
   fi
   SLURM_NPROCS=`expr ${SLURM_JOB_NUM_NODES} \* ${SLURM_NTASKS_PER_NODE}`
fi

PREMANAGERCOMMANDS
MANAGERCOMMAND `pwd`/APPSCRIPTNAME > JOBID.stdout 2> JOBID.stderr
POSTMANAGERCOMMANDS

date +"%s" > TS_FINISH

touch TIME_RESULTS-0
cat TIME_RESULTS-[0-9]* >> TIME_RESULTS
rm -f TIME_RESULTS-[0-9]*
"""


   def __buildMPIFile(self):
      # setup regex's for the template
      re_preManagerCommands  = re.compile("PREMANAGERCOMMANDS")
      re_managerCommand      = re.compile("MANAGERCOMMAND")
      re_postManagerCommands = re.compile("POSTMANAGERCOMMANDS")
      re_wallTime            = re.compile("WALLTIME")
      re_nnodes              = re.compile("NNODES")
      re_ppn                 = re.compile("PPN")
      re_processors          = re.compile("NPROCESSORS")
      re_partition           = re.compile("PARTITION")
      re_constraints         = re.compile("CONSTRAINTS")
      re_environment         = re.compile("ENVIRONMENT")
      re_appScriptName       = re.compile("APPSCRIPTNAME")
      re_tsStart             = re.compile("TS_START")
      re_tsFinish            = re.compile("TS_FINISH")
      re_timeResults         = re.compile("TIME_RESULTS")
      re_jobid               = re.compile("JOBID")

      template = self.__makeMPITemplate()

      partitionSBATCH = ""
      if self.remoteBatchPartition != "":
         partitionSBATCH = "#SBATCH --partition " + self.remoteBatchPartition

      constraintSBATCH = ""
      if self.remoteBatchConstraints != "":
         constraintSBATCH = "#SBATCH --constraint " + self.remoteBatchConstraints

      environmentExport = ""
      environmentVars = self.environment.split()
      for environmentVar in environmentVars:
         environmentExport += "export " + environmentVar + "\n"

      template = re_preManagerCommands.sub("\n".join(self.preManagerCommands),template)
      template = re_postManagerCommands.sub("\n".join(self.postManagerCommands),template)
      template = re_managerCommand.sub(self.managerCommand,template)
      template = re_wallTime.sub(self.wallTime,template)
      if self.remoteBatchPartitionSize != "":
         template = re_nnodes.sub(self.remoteBatchPartitionSize,template)
      else:
         template = re_nnodes.sub(self.nNodes,template)
      template = re_ppn.sub(self.ppn,template)
      nProcessors = str(int(self.nNodes)*int(self.ppn))
      template = re_processors.sub(nProcessors,template)
      template = re_partition.sub(partitionSBATCH,template)
      template = re_constraints.sub(constraintSBATCH,template)
      template = re_environment.sub(environmentExport,template)
      template = re_appScriptName.sub(self.appScriptName,template)
      template = re_tsStart.sub(self.timestampStart,template)
      template = re_tsFinish.sub(self.timestampFinish,template)
      template = re_timeResults.sub(self.timeResults,template)
      template = re_jobid.sub("%s_%s" % (self.localJobId,self.instanceId),template)

      return(template)


   def __makeMPINoForkTemplate(self):
      return """#!/bin/sh
PARTITION
CONSTRAINTS
#SBATCH --nodes=NNODES
#SBATCH --ntasks-per-node=PPN
#SBATCH --time=WALLTIME
#SBATCH -J JOB_JOBID
#SBATCH -o slurm_JOBID.stdout
#SBATCH -e slurm_JOBID.stderr

ENVIRONMENT

# RemoteBatchSLURM:makeMPINoForkTemplate

trap cleanup HUP INT QUIT ABRT TERM

cleanup()
{
   echo "Abnormal termination by signal"
   if [ ! -s TS_FINISH ] ; then
      date +"%s" > TS_FINISH
   fi
   exit 1
}

export PATH=${SLURM_SUBMIT_DIR}:${PATH}
# Change to directory where job was submitted.
cd ${SLURM_SUBMIT_DIR}

date +"%s" > TS_START

PREMANAGERCOMMANDS
MANAGERCOMMAND EXECUTABLE ARGUMENTS STDIN > JOBID.stdout 2> JOBID.stderr
POSTMANAGERCOMMANDS

date +"%s" > TS_FINISH

startTime=`cat TS_START`
finishTime=`cat TS_FINISH`
let wallTime=finishTime-startTime
cnt=0
while [ "${cnt}" != "NPROCESSORS" ] ; do
   cat >> TIME_RESULTS <<EOF
real ${wallTime}.00
user ${wallTime}.00
sys 0.00
EOF
   cnt=`expr ${cnt} + 1`
done
"""


   def __buildMPINoForkFile(self):
      # setup regex's for the template
      re_stdinput            = re.compile("STDIN")
      re_executable          = re.compile("EXECUTABLE")
      re_arguments           = re.compile("ARGUMENTS")
      re_preManagerCommands  = re.compile("PREMANAGERCOMMANDS")
      re_managerCommand      = re.compile("MANAGERCOMMAND")
      re_postManagerCommands = re.compile("POSTMANAGERCOMMANDS")
      re_wallTime            = re.compile("WALLTIME")
      re_nnodes              = re.compile("NNODES")
      re_ppn                 = re.compile("PPN")
      re_processors          = re.compile("NPROCESSORS")
      re_partition           = re.compile("PARTITION")
      re_constraints         = re.compile("CONSTRAINTS")
      re_environment         = re.compile("ENVIRONMENT")
      re_tsStart             = re.compile("TS_START")
      re_tsFinish            = re.compile("TS_FINISH")
      re_timeResults         = re.compile("TIME_RESULTS")
      re_jobid               = re.compile("JOBID")

      template = self.__makeMPINoForkTemplate()

      partitionSBATCH = ""
      if self.remoteBatchPartition != "":
         partitionSBATCH = "#SBATCH --partition " + self.remoteBatchPartition

      constraintSBATCH = ""
      if self.remoteBatchConstraints != "":
         constraintSBATCH = "#SBATCH --constraint " + self.remoteBatchConstraints

      environmentExport = ""
      environmentVars = self.environment.split()
      for environmentVar in environmentVars:
         environmentExport += "export " + environmentVar + "\n"

      if self.stdinput == "":
         template = re_stdinput.sub(self.stdinput,template)
      else:
         template = re_stdinput.sub(" < "+self.stdinput,template)
      template = re_executable.sub(self.executable,template)
      template = re_arguments.sub(self.arguments,template)
      template = re_preManagerCommands.sub("\n".join(self.preManagerCommands),template)
      template = re_postManagerCommands.sub("\n".join(self.postManagerCommands),template)
      template = re_managerCommand.sub(self.managerCommand,template)
      template = re_wallTime.sub(self.wallTime,template)
      if self.remoteBatchPartitionSize != "":
         template = re_nnodes.sub(self.remoteBatchPartitionSize,template)
      else:
         template = re_nnodes.sub(self.nNodes,template)
      template = re_ppn.sub(self.ppn,template)
      nProcessors = str(int(self.nNodes)*int(self.ppn))
      template = re_processors.sub(nProcessors,template)
      template = re_partition.sub(partitionSBATCH,template)
      template = re_constraints.sub(constraintSBATCH,template)
      template = re_environment.sub(environmentExport,template)
      template = re_tsStart.sub(self.timestampStart,template)
      template = re_tsFinish.sub(self.timestampFinish,template)
      template = re_timeResults.sub(self.timeResults,template)
      template = re_jobid.sub("%s_%s" % (self.localJobId,self.instanceId),template)

      return(template)


   def buildBatchScript(self):
      batchLogName = ""
      batchScriptName = "%s_%s.slurm" % (self.localJobId,self.instanceId)
      if self.isMultiCoreRequest:
         if   self.computationMode == 'mpi':
            batchScript = self.__buildMPIFile()
         elif self.computationMode == 'mpiNoFork':
            batchScript = self.__buildMPINoForkFile()
         elif self.computationMode == 'parallel':
            batchScript = self.__buildSerialFile()
      else:
         batchScript = self.__buildSerialFile()

      return(batchLogName,batchScriptName,batchScript)


   def getBatchNodeList(self):
      return(self.nodeFileName,self.nodeList)


